/***************************************************************************
*
* Authors:     Carlos Oscar S. Sorzano (coss@cnb.csic.es)
*
* Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
* 02111-1307  USA
*
*  All comments concerning this program package may be sent to the
*  e-mail address 'xmipp@cnb.csic.es'
***************************************************************************/

#ifndef ARGS_H
#define ARGS_H

#include <cstdio>
#include <string>
#include <vector>
//ROB
#include <string.h>

#include "funcs.h"
template <typename T> class Matrix1D;

/** @defgroup Arguments Arguments parsing
 *  @ingroup DataLibrary
 *
 * This set of functions are designed for make easier checking and reading
 * parameters from a string. The single value readings don't modify the value of
 * the input string, while the multiple value readings do. In general, the
 * reading is oriented to be done with a chain of strtok (a general C function),
 * for this reason the string itself is modified by the succesive calls to
 * strtok.
 *
 * The second group (list reading) uses tokens internally, while in the single
 * read functions the token must be given by hand. Anyway, it is not so
 * difficult to read a set of fields from a line with all the normal checks we
 * would like (existence of a parameter, checking for numerical correctness...)
 *
 * The following is an example of its use. This code tries to read a number, a
 * character and a list of numbers from a line. In the example you can also see
 * how to write code for the different error handling methods.
 *
 * @code
 * std::string line;
 * int key, param_no;
 * std::vector< float > data;
 *
 * #ifndef _NO_EXCEPTION
 * try
 * {
 * #endif
 *
 * key = textToFloat(firstToken(line), 1602, "Error reading key");
 * param_no = textToInt(nextToken(), 1602, "Error reading number parameters");
 * readFloatList(NULL, param_no, data, 1602, "Error reading doc file line");
 *
 * #ifndef _NO_EXCEPTION
 * }
 * catch (Xmipp_error XE)
 * {
 *     std::cout << XE;
 *     DL.line_type = 0;
 *     REPORT_ERROR(1602, "Line is discarded");
 * }
 * #endif
 *
 * @endcode
 */

/** @defgroup TypeConversions Type conversions
 * @ingroup Arguments
 *
 * All these functions try to produce a value of the desired type from the given
 * string (char*) or to produce a string from a value. If it is not possible
 * (there is nothing in the input string, there's something which is not a
 * float value), then the function throws an exception. There are default values
 * for the exception thrown but you can provide a different exception in the
 * parameter line
 */

/** Best precision for a float number.
 * @ingroup TypeConversions
 *
 * This function returns the best precision to be used in a "printf" format if
 * this number is to fit in a given width. It returns -1 if the exponential
 * format is advised.
 *
 * @code
 * template<typename T>
 * std::ostream& operator<<(std::ostream& out, const T& val)
 * {
 *     int i,j;
 *
 *     if (val.xdim == 0)
 *         out << "NULL matrix" << std::endl;
 *     else
 *         out << std::endl;
 *
 *     T aux = ABSnD(val);
 *     int prec = bestPrecision(aux.max(), 10);
 *
 *     for (i=STARTINGY(val); i<=FINISHINGY(val); i++)
 *     {
 *         for (j=STARTINGX(val); j<=FINISHINGX(val); j++)
 *         {
 *             out << floatToString((float) val(i,j), 10, prec) << ' ';
 *         }
 *         out << std::endl;
 *     }
 *
 *     return out;
 * }
 *
 * @endcode
 */
int bestPrecision(float F, int _width);

/** String (char*) to double conversion.
 * @ingroup TypeConversions
 *
 * @code
 * double key = textToDouble(firstToken(line), 1602, "Error reading key");
 * @endcode
 */
double textToDouble(const char* str,
            int _errno = 2101,
            std::string errmsg = "Error in textToDouble",
            int exit = 0);

/** String (char*) to float conversion.
 * @ingroup TypeConversions
 *
 * @code
 * float key = textToFloat(firstToken(line), 1602, "Error reading key");
 * @endcode
 */
float textToFloat(const char* str,
           int _errno = 2101,
           std::string errmsg = "Error in textToFloat",
           int exit = 0);

/** String (STL) to float conversion.
 * @ingroup TypeConversions
 *
 * @code
 * float key = textToFloat(str, 1602, "Error reading key");
 * @endcode
 */
inline float textToFloat(const std::string str,
                  int _errno = 2101,
                  std::string errmsg = "Error in textToFloat",
                  int exit = 0)
{
    return textToFloat(str.c_str(), _errno, errmsg, exit);
}

/** String (char*) to integer conversion.
 * @ingroup TypeConversions
 *
 * @code
 * int param_no = textToInteger(nextToken(), 1602, "Error reading number parameters")
 * @endcode
 */
int textToInteger(const char* str,
         int _errno = 2102,
         std::string errmsg = "Error in textToInteger",
         int exit = 0);

/** String (STL) to integer conversion.
 * @ingroup TypeConversions
 *
 * @code
 * int param_no = textToInteger(str, 1602, "Error reading number parameters")
 * @endcode
 */
inline int textToInteger(const std::string str,
                int _errno = 2102,
                std::string errmsg = "Error in textToInteger",
                int exit = 0)
{
    return textToInteger(str.c_str(), _errno, errmsg, exit);
}

/** String (char*) to long long integer conversion.
 * @ingroup TypeConversions
 *
 * @code
 * long long param_no = textToLongLong(nextToken(), 1602, "Error reading number
 *     parameters")
 * @endcode
 */
long long textToLongLong(const char* str,
                int _errno = 2102,
                std::string errmsg = "Error in AtoL",
                int exit = 0);

/** Float to string conversion.
 * @ingroup TypeConversions
 *
 * If precision==0 the precision is automatically computed in such a way that
 * the number fits the width (the exponential format might be chosen). If
 * precision==-1 then the exponential format is forced. If width==0 then the
 * minimum width is used.
 *
 * @code
 * REPORT_ERROR(1602, "Value not recognised " + floatToString(val));
 * @endcode
 */
std::string floatToString(float F, int _width = 8, int _prec = 0);

/** Integer to string conversion.
 * @ingroup TypeConversions
 *
 * If width==0 then writes the number with the number of digits needed. The
 * fill_with field indicates which is the filling character for the left
 * positions.
 *
 * @code
 * REPORT_ERROR(1602, "Error reading key " + integerToString(key));
 * @endcode
 */
std::string integerToString(int I, int _width = 0, char fill_with = '0');

/** Character to integer conversion.
 * @ingroup TypeConversions
 *
 * Takes a character and produces a number according to its ASCII code minus 48.
 * For instance, ASCII=48 produces number 0, ASCII=49 produces 1, ..., ASCII=57
 * produces 9, ASCII=58 produces 10!!, ... This is used when you have codified
 * numbers greater than 9 in a single character.
 *
 * @code
 * int param_no = textToInt(token, 1602, "Error reading number parameters");
 * @endcode
 */
int textToInt(const char* str,
         int _errno = 2103,
         std::string errmsg = "Error in textToInt",
         int exit = 0);

/** String to string with given length conversion.
 * @ingroup TypeConversions
 *
 * The output string will have the information of the input one with the given
 * width. If the width is smaller than the string length then the string is
 * truncated and if it is greater the string is right padded with spaces. If
 * width==0 then the same string is returned.
 */
std::string stringToString(const std::string& str, int _width = 0);

/** Check angle.
 * @ingroup TypeConversions
 *
 * If the argument is not "rot", "tilt" nor "psi" an exception is thrown
 */
void checkAngle(const std::string& str);

/** To lower.
 * @ingroup TypeConversions
 *
 * All characters between A-Z are brought to a-z. Result is rewritten on input
 * string
 */
void toLower(char* _str);

/** To lower, for STL strings.
 * @ingroup TypeConversions
 */
void toLower(std::string& _str);

/** Remove consecutive spaces.
 * @ingroup TypeConversions
 *
 * All consecutive spaces are replaced by a single one and starting and
 * finishing spaces are removed
 */
std::string removeSpaces(const std::string& _str);

/** Remove quotes.
 * @ingroup TypeConversions
 *
 * This function removes the first character if it is a double or single quote,
 * as well as the last character. The char pointer might be moved.
 *
 * @code
 * char str[10] = "\"Hello\"";
 * (&str);
 * @endcode
 */
void removeQuotes(char** _str);

/** @defgroup Tokenization Tokenization
 * @ingroup Arguments
 *
 * These functions allow to split a string into small pieces separated by blank
 * spaces, giving you a pointer to the different word each time. The different
 * elements from the string are selected using strtok, so after the application
 * of this function to the input string, this is modified and NULL characters
 * are introduced as delimiters of the elements. This is useful in most
 * situations since after reading a list you might go on reading more things,
 * but you must be aware of it. Here goes an example of doing so:
 *
 * @code
 * std::cout << "Whole  line: " << line << std::endl;
 * std::cout << "First  word: " << firstToken(line) << std::endl;
 * std::cout << "Second word: " << nextToken() << std::endl;
 * std::cout << "Third  word: " << nextToken() << std::endl;
 * ...
 * @endcode
 *
 * When there are no more words, both functions return a NULL pointer. Here we
 * make a distinction between tokens (words that might be empty) and words
 * (words that cannot be empty, if they are then an exception or an exit error
 * is thrown).
 *
 * For STL there is another way. You supply a string object and a vector of
 * strings is returned with all the elements
 */

/** Split a STL string given some delimiter.
 * @ingroup Tokenization
 *
 * Returns a the number of tokens found. The tokens are in the variable results.
 */
int splitString(const std::string& input,
                const std::string& delimiter,
                std::vector< std::string >& results,
                bool includeEmpties = false);

/** Returns first token (char*).
 * @ingroup Tokenization
 *
 * @code
 * char line[80];
 *
 * std::cout << "First  word: " << firstToken(line) << std::endl;
 * @endcode
 */
inline char* firstToken(const char* str)
{
    return strtok((char*) str, " \t\n");
}

/** Returns first token (STL).
 * @ingroup Tokenization
 *
 * @code
 * std::string line;
 *
 * std::cout << "First  word: " << firstToken(line) << std::endl;
 * @endcode
 */
inline char* firstToken(const std::string& str)
{
    // FIXME C-style cast
    return strtok((char*) str.c_str(), " \t\n");
}

/** Returns next token.
 * @ingroup Tokenization
 *
 * This functions returns the next word of the line we have given last as
 * parameter to firstToken.
 *
 * @code
 * char line[80];
 * ...
 * firstToken(line);
 * std::cout << "Second  word: " << nextToken(line) << std::endl;
 *
 * stf::string line;
 * ...
 * firstToken(line);
 * std::cout << "Second  word: " << nextToken(line) << std::endl;
 * @endcode
 */
inline char* nextToken()
{
    // FIXME C-style cast
    return strtok((char*) NULL, " \t\n");
}

/** Returns next token.
 * @ingroup Tokenization
 *
 * It reads from position i. Returns (in i) the following position to search on.
 * When there are no more tokens. It returns "".
 */
std::string nextToken(const std::string& str, int& i);

/** Get non empty string (char*).
 * @ingroup Tokenization
 *
 * This function returns the first word found in the given line disregarding the
 * leading blanks. If no word is found then an exception or an exit error is
 * produced. After calling this function the first blank character after the
 * word is substituted by a NULL character (as it uses the function firstToken.
 * Further word readings should use the function read_nextWord
 */
char* firstWord(char* str,
                 int _errno = 2106,
                 const std::string & errmsg = "first word: String not found",
                 int exit = 0);

/** Get non empty string (STL).
 * @ingroup Tokenization
 *
 * Same as the previous function but for STL strings
 */
inline char* firstWord(std::string& str,
                        int _errno = 2106,
                        std::string errmsg = "first word: String not found",
                        int exit = 0)
{
    // FIXME C-style cast
    return firstWord((char*) str.c_str(), _errno, errmsg, exit);
}

/** Get next non empty string.
 * @ingroup Tokenization
 *
 * This is the same as the nextToken, but an exception is thrown or an exit
 * error produced if the word is empty
 */
inline char* nextWord(int _errno = 2106,
                       std::string errmsg = "next word: String not found",
                       int exit = 0)
{
    // FIXME C-style cast
    return firstWord((char*) NULL, _errno, errmsg, exit);
}

// TODO Document
/// @ingroup Tokenization
void tokenize(const std::string& str,
              std::vector< std::string >& tokens,
              const std::string& delimiters = " \t");

/** @defgroup ReadLists Read lists
 * @ingroup Arguments
 *
 * These functions try to read N values of the desired type into the given
 * structure (either a STL vector of any numerical type by adding the read
 * values at the end or a Matrix1D of any numerical type and then the values are
 * written at PHYSICAL positions 0 ... N-1, the Matrix1D must be big enough to
 * hold all the data since it is not resized internally.
 *
 * If it is not possible to read all parameters an exception is thrown. You can
 * provide the exception in the function call but there are default values.
 * These functions are based in those for reading single values, so it might be
 * possible that these other functions throw other exceptions.
 *
 * The different elements of the list are selected using the tokenizing
 * functions (different elements in the string are separated by spaces), so
 * after the application of this function to the input string, this is modified
 * and NULL characters are introduced as delimiters of the elements.
 *
 * The following code is an example of doing so:
 *
 * @code
 * getline(in_stream, line);
 * readFloatList(line, 10, v1); // Read 10 values from line in v1
 * readFloatList(NULL, 10, v2); // Read NEXT!! 10 values in v2
 * @endcode
 */

/** List to STL vector.
 * @ingroup ReadLists
 */
template <typename T>
void readFloatList(const char* str,
                     int N, std::vector< T >& v,
                     int _errno = 2105,
                     std::string errmsg = "Error reading list",
                     int exit = 0)
{
    T valueF;
    char* token;

    token = firstToken(str);
    for (int i = 0; i < N; i++)
    {
        if (token == NULL)
        {
            // CO: Should not report other error than the required one
            // std::cout << "Read float list: Number of true parameters doesn't \
            // coincide\n";
            REPORT_ERROR(_errno, errmsg);
        }

        try
        {
            valueF = (T) textToFloat(token);
        }
        catch (Xmipp_error)
        {
            REPORT_ERROR(_errno, errmsg);
        }

        v.push_back(valueF);

        if (i != N - 1)
            token = nextToken();
    }
}

/** List to STL vector.
 * @ingroup ReadLists
 */
template <typename T>
void readFloatList(const std::string& str,
                     int& i,
                     int N,
                     std::vector< T >& v,
                     int _errno = 2105,
                     std::string errmsg = "Error reading list",
                     int exit = 0)
{
    T valueF;
    std::string token;

    token = nextToken(str, i);
    for (int j = 0; j < N; j++)
    {
        if (token == "")
        {
            // CO: Should not report other error than the required one
            // std::cout << "Read float list: Number of true parameters doesn't \
            // coincide\n";
            REPORT_ERROR(_errno, errmsg);
        }

        try
        {
            valueF = (T) textToFloat(token.c_str());
        }
        catch (Xmipp_error)
        {
            REPORT_ERROR(_errno, errmsg);
        }

        v.push_back(valueF);

        if (j != N - 1)
            token = nextToken(str, i);
    }
}

/** List to Matrix1D.
 * @ingroup ReadLists
 */
template <typename T>
void readFloatList(const char* str,
                     int N,
                     Matrix1D< T >& v,
                     int _errno = 2105,
                     std::string errmsg = "Error reading floating list",
                     int exit = 0)
{
    T valueF;
    char* token;

    token = firstToken(str);
    for (int i = 0; i < N; i++)
    {
        if (token == NULL)
        {
            // CO: Should not report other error than the required one
            // std::cout << "Read float list: Number of true parameters doesn't \
            // coincide\n";
            REPORT_ERROR(_errno, errmsg);
        }

        try
        {
            valueF = (T) textToFloat(token);
        }
        catch (Xmipp_error)
        {
            REPORT_ERROR(_errno, errmsg);
        }

        DIRECT_VEC_ELEM(v, i) = valueF;
        if (i != N - 1)
            token = nextToken();
    }
}

/** @defgroup CommandLineFunctions Functions for parsing the command line
 * @ingroup Arguments
 *
 * These functions help you to manage the command line parameters
 */

/** Get parameters from the command line.
 * @ingroup CommandLineFunctions
 *
 * This function assumes that the command line is structured in such a way that
 * for each parameter a block like "-param <param_value>" is defined. The label
 * "param" can be substituted by any other one you like. If the parameter is
 * optional then this function allows you to define a default value. If no
 * default value is supplied and the parameter is not specified in the command
 * line, then an exception is thrown. You may change the default exception.
 *
 * You may also indicate that in case of error no exception is raised and force
 * the program to abort (use the exit variable).
 *
 * @code
 * m_param = textToFloat(getParameter(argc, argv, "-m"));
 *
 * // Get compulsory parameter "-m"
 * m_param = textToFloat(getParameter(argc, argv, "-m","2.65"));
 *
 * // Optional parameter, if no parameter is given it takes 2.65 by default
 * m_param = textToFloat(getParameter(argc, argv, "-m", NULL, 6001, "-m parameter not \
 *     found. I'm going out", TRUE);
 *
 * // Compulsory parameter, if not found give an special error message and exit
 * // the program
 *
 * @endcode
 */
char* getParameter(int argc,
                char** argv,
                const char* param,
                const char* option = NULL,
                int _errno = -1,
                const std::string & errmsg = "",
                int exit = 0);

/** Get two float parameters after a flag from the command line.
 * @ingroup CommandLineFunctions
 *
 * An exception is thrown if there are not enough parameters after the flag, if
 * the message is empty then "Not enough parameters after <param>" is shown. The
 * default values must be provided. TRUE is returned if the two values have been
 * found
 */
bool getTwoDoubleParams(int argc,
                         char** argv,
                         const char* param,
                         double& v1,
                         double& v2,
                         double v1_def,
                         double v2_def,
                         int _errno = 2104,
                         const std::string & errmsg = "",
                         int exit = 0);

/** Get 3 float parameters after a flag from the command line.
 * @ingroup CommandLineFunctions
 *
 * An exception is thrown if there are not enough parameters after the flag, if
 * the message is empty then "Not enough parameters after <param>" is shown. The
 * default values must be provided. TRUE is returned if the two values have been
 * found
 */
bool getThreeDoubleParams(int argc,
                         char** argv,
                         const char* param,
                         double& v1,
                         double& v2,
                         double& v3,
                         double v1_def,
                         double v2_def,
                         double v3_def,
                         int _errno = 2104,
                         const std::string& errmsg = "",
                         int exit = 0);

/** Get boolean parameters from the command line.
 * @ingroup CommandLineFunctions
 *
 * This function assumes that the command line is structured in such a way that
 * for each parameter a block like "-param" is defined. The label "param" can be
 * substituted by any other one you like. It might be used to look for a boolean
 * parameter, for instance:
 *
 *     -verbose means that verbose functionality is set (TRUE)
 *
 * @code
 * verbose = checkParameter(argc, argv, "-verbose"));
 *
 * // checks if "-verbose" was supplied in the command line. If -verbose was
 * // supplied the function returns TRUE (1), otherwise returns FALSE (0)
 * @endcode
 */
bool checkParameter(int argc, char** argv, const char* param);

/** Returns the position where the given parameter is in the command line.
 * @ingroup CommandLineFunctions
 *
 * This function assumes that the command line is structured in such a way that
 * for each parameter a block like "-param" is defined. The label "param" can be
 * substituted by any other one you like. It returns -1 if the parameter is not
 * found. It is used to look for parameters having a list of values behind, for
 * instance:
 *
 *     -ang rot tilt psi
 *
 * @code
 * i = paremeterPosition(argc, argv, "-ang"));
 *
 * // This condition checks if 3 arguments where introduced after -ang parameter
 * // (assuming that the -ang argument is the last one in the string)
 * if (i+3 >= argc)
 *     EXIT_ERROR(1, "Not enough parameters behind -ang");
 *
 * ang1 = argv[i+1];
 * ang2 = argv[i+2];
 * ang3 = argv[i+3];
 * @endcode
 */
int paremeterPosition(int argc, char** argv, const char* param);


/** Return the number of components of a vector argument.
 * @ingroup CommandLineFunctions
 *
 * A vector argument is defined as [x,y,z,...]. It returns 0 if the string does
 * not contain a vector
 */
int numComponents(const std::string& str);

/** Get float vector.
 * @ingroup CommandLineFunctions
 *
 * A vector is defined as a "[x,y,z, ...]" of any dimension (by default 2D
 * vectors are read). The vector must not contain blank spaces.
 *
 * @code
 * a = getVectorParameter(argc, argv, "-a", 3); // a 3D vector called "-a".
 * @endcode
 *
 * The vector is internally resized properly. If the dimension is -1 then all
 * the vector components are read disregarding their dimensionality, ie,-1 is
 * used to read vectors of an unknown dimension. If the parameter is not found
 * when no dimensionality is given an empty vector is returned but no exception
 * is thrown. If there is no dimensionality and a single parameter is behind the
 * flag then no brackets are needed
 */
Matrix1D< double > getVectorParameter(int argc,
                                    char** argv,
                                    const char* param,
                                    int dim = 2,
                                    int _errno = -1,
                                    const std::string & errmsg = "",
                                    int exit = 0);

/** Get specific command line.
 * @ingroup CommandLineFunctions
 *
 * This function allows to pass command line parameters to several programs
 * which are integrated in a single program. It assumes that the command line is
 * a compound line like this
 *
 * @code
 * whole_process prog1 ( -i in_file1 -o out_file1 ) prog2 ( -i out_file1 )
 * @endcode
 *
 * In this example, a single program "whole_process" is compund of 2 programs
 * ("prog1" and "prog2"). The parameters for prog1 are within the brackets after
 * prog1 (notice that spaces around brackets are very important), and then the
 * output of program 1 is the input of program 2.
 *
 * This function returns new argc and argv such that prog1 parameters could be
 * read from this new command line.
 *
 * @code
 * int argcp;
 * char** argvp;
 *
 * specificCommandLine("prog1", argc, argv, argcp, &argvp);
 *
 * if (argcp == 0)
 *     EXIT_ERROR(1, "No parameters for prog1");
 *
 * prog1_parameters.read(argcp, argvp);
 * @endcode
 *
 * If the program name is not found or the corresponding brackets then argcp==0
 */
void specificCommandLine(const std::string& prog_name,
                           int argc,
                           char** argv,
                           int& argcp,
                           char*** argvp);

/** Generate argc and argv for a string.
 * @ingroup CommandLineFunctions
 *
 * Given a string this function makes a copy of the string and divides it into
 * tokens such that they can be used as argc and argv, as if it were a command
 * line.
 *
 * The string copy remains in "copy" and it can be freed by disposing this
 * variable.
 *
 * argvp[0] (normally the program name) is set to any value, in this case to
 * "autom", standing for "automatically generated".
 *
 * argcp==0 if now valid command line is provided, ie, the line is empty or only
 * with blanks.
 *
 * Next time the function is called it checks that argv and copy are empty
 * (pointing to NULL), if they aren't then firstly the associated memory is
 * freed.
 *
 * @code
 * int argcp;
 * char** argvp;
 * char* copy;
 *
 * copy = NULL;
 * argvp = NULL;
 *
 * string command_line = "-i input_file -o output_file";
 *
 * generateCommandLine(command_line, argcp, &argvp, &copy);
 *
 * if (argcp != 0)
 *     read_parameters(argcp, argvp);
 * @endcode
 */
void generateCommandLine(const std::string& command_line,
                           int& argcp,
                           char**& argvp,
                           char*& copy);

/** Generate articial command line from a file.
 * @ingroup CommandLineFunctions
 *
 * The copy variable must be destroyed outside by "delete copy". This function
 * takes "input_file=<input_file>" and turns it into "-input_file <input_file>"
 * The appropiate argc, argv are also returned.
 *
 * Returns TRUE if the parameter is found in the file, and FALSE if it is not
 */
bool generateCommandLine(FILE* fh,
                           const char* param,
                           int& argcp,
                           char**& argvp,
                           char*& copy);

/** Get parameter from file.
 * @ingroup CommandLineFunctions
 *
 * Parameters are supposed to be identified with an =, so any line which doesn't
 * contain an = character cannot contain a parameter. The left part of the "="
 * is the identifier, and the right part the value. The value is returned
 * without any extra space, if it is compound of several words, then the spaces
 * in between are simplified to a single blank space.
 *
 * The file position inside the file is not moved and comments are allowed
 * starting with "#" and ";".
 *
 * Parameter skip controls the number of these parameters to skip before
 * returning the value, ie, if there are several "parameter=" tags in a file and
 * you want the first one then you should skip 0, if you want the second the
 * skip=1, ...
 *
 * The meaning and use of the exit, errors and optional value is the same as in
 * the command line getParameter
 */
std::string getParameter(FILE* fh,
                      const char* param,
                      int skip = 0,
                      const char* option = NULL,
                      int _errno = -1,
                      const std::string & errmsg = "",
                      int exit = 0);

/** Check if a parameter is present in a file.
 * @ingroup CommandLineFunctions
 *
 * The same as the previous function, but this function only reports if a
 * parameter is present or not in a file. Notice that boolean parameters must be
 * defined as "parameter=". If after the parameter, "no" comes then this
 * function returns FALSE
 */
bool checkParameter(FILE* fh, const char* param);

/** Get float vector from a file.
 * @ingroup CommandLineFunctions
 *
 * The same as before but reading is done from a file
 */
Matrix1D< double > getVectorParameter(FILE* fh,
                                    const char* param,
                                    int dim = 2,
                                    int _errno = -1,
                                    const std::string & errmsg = "",
                                    int exit = 0);

#endif
