/***************************************************************************
 *
 * Authors:     Carlos Oscar S. Sorzano (coss@cnb.csic.es)
 *              Roberto Marabini (roberto@mipg.upenn.edu)
 *
 * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 * 02111-1307  USA
 *
 *  All comments concerning this program package may be sent to the
 *  e-mail address 'xmipp@cnb.csic.es'
 ***************************************************************************/
/*****************************************************************************/
/* APH Files: MRC                                                            */
/*****************************************************************************/

#ifndef _XMIPP_APH_HH
#define _XMIPP_APH_HH

#include <data/funcs.h>
#include <data/matrix2d.h>
#include <data/geometry.h>

/**@defgroup APH2D APH2D Files
   @ingroup InterfaceLibrary */
//@{
/** APH2D Files.
    This is a class to read/write APH files as produced by the MRC's program
    mmbox.
    @code
1410065407   24.8900    0.0000    0.0000   16.0000  1024  1024   2.500
 -31 -31     4.0    34.0  1     0.0     0.0
 -31 -30     4.0   337.0  1     0.0     0.0
 -31 -29     4.0   280.0  1     0.0     0.0
 -31 -28     4.0   224.0  1     0.0     0.0
 -31 -27     4.0   169.0  1     0.0     0.0
    @endcode
    The first line used to be skipped. Now, it is
    \\(???) (a*x) (a*y) (b*x) (b*y) (crystal_Xdim) (crystal_Ydim) (sampling rate)

    Each line describes a Fourier spot, the first
    two columns are h and k, then the magnitude, the phase, the spot IQ,
    spot background, finally the last term is related with the CTF correction
    */
class APHFile2D
{
public:
    /** APH Filename */
    FileName fn_aph;
    /** Label */
    int label;
    /** a* vector */
    Matrix1D<double> astar;
    /** b* vector */
    Matrix1D<double> bstar;
    /** Crystal X dimensions */
    int Xdim;
    /** Crystal Y dimensions */
    int Ydim;
    /** Sampling rate */
    double sampling_rate;
    /** Matrices with the FT spots. */
    Matrix2D<double> spots_abs;
    Matrix2D<double> spots_arg;
    /** Matrix with the spots IQ. */
    Matrix2D<int> IQ;
    /** Matrix with the  spots Background. */
    Matrix2D<double> background;
    /** Matrix with the  spots CTF. */
    Matrix2D<double> CTF;
    /** Matrix with the  spots height (l, z*). */
    Matrix2D<double> spots_l;
public:
    // l is present in the read file
    bool l_is_present;
public:
    /** Read an aph file generated by the MRC program mmbox.
        first line is saved as comment in the header. 5 matrices are created
        containing the magnitude and phase (spots_abs abd spots_arg), the IQs (IQ),      the background (background) and finally the CTF related stuff (CTF).
        */
    void read(const FileName &fn);

    /** Write the content of this object to a file.*/
    void write(const FileName &fn) const;

    /** Empties actual APH structure. */
    void clear();

    /** Copy reflection. Copy a reflection from (h,k) to (new_h,new_k). The
        associated information is also copied (IQ, background and CTF). The sign of
        the new reflection can be modified through sign. If sign is an even number,
        no sign change is performed. Otherwise, the sign of the spot is changed (a
        phase change of 180 degrees).

        The phase of the original reflection can also be conjugated in order to
        produce the new spot.*/
    void copy_reflection(int h, int k, int new_h, int new_k, double new_l,
                         bool conjugate, int sign);
#ifdef DISCONTINUATED
    /** Generate symmetrical spot reflections.
        See the symmetries.hh file in the Reconstruction library to know
        about valid symmetry groups.*/
    void generate_symmetrical_reflections(int symmetry_group);
#endif
};
/**  Transform Xmipp Euler angles into MRC angles  */
void Euler_to_MRC(double rot, double tilt, double psi,
                  double *mrc_tilt, double *mrc_taxa);
/**  Transform Xmipp MRC angles into EULER angles, psi is set to 0  */
void MRC_to_Euler(double mrc_taxa, double mrc_tilt,
                  double *rot, double *tilt, double *psi);

//@}
#endif
