/***************************************************************************
 *
 * Authors:
 *
 * Carlos Oscar S. Sorzano (coss@cnb.csic.es)
 * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 * 02111-1307  USA
 *
 *  All comments concerning this program package may be sent to the
 *  e-mail address 'xmipp@cnb.csic.es'
 ***************************************************************************/

#include "convert_pdb2vol.h"

#include <data/args.h>

#include <fstream>

/* Empty constructor ------------------------------------------------------- */
Prog_PDBPhantom_Parameters::Prog_PDBPhantom_Parameters()
{
    blob.radius = 2;   // Blob radius in voxels
    blob.order  = 2;   // Order of the Bessel function
    blob.alpha  = 3.6; // Smoothness parameter
    output_dim = -1;
    fn_pdb = "";
    Ts = 1;
    highTs = 1.0/12.0;
    useBlobs=false;
    usePoorGaussian=false;

    // Periodic table for the blobs
    periodicTable.resize(7, 2);
    periodicTable(0, 0) = atomRadius("H");
    periodicTable(0, 1) = atomCharge("H");
    periodicTable(1, 0) = atomRadius("C");
    periodicTable(1, 1) = atomCharge("C");
    periodicTable(2, 0) = atomRadius("N");
    periodicTable(2, 1) = atomCharge("N");
    periodicTable(3, 0) = atomRadius("O");
    periodicTable(3, 1) = atomCharge("O");
    periodicTable(4, 0) = atomRadius("P");
    periodicTable(4, 1) = atomCharge("P");
    periodicTable(5, 0) = atomRadius("S");
    periodicTable(5, 1) = atomCharge("S");
    periodicTable(6, 0) = atomRadius("Fe");
    periodicTable(6, 1) = atomCharge("Fe");
    // Correct the atom weights by the blob weight
    for (int i = 0; i < YSIZE(periodicTable); i++)
    {
        periodicTable(i, 1) /=
            basvolume(periodicTable(i, 0) / highTs, blob.alpha, blob.order, 3);
    }
}

/* Produce Side Info ------------------------------------------------------- */
void Prog_PDBPhantom_Parameters::produceSideInfo() {
    if (!useBlobs && !usePoorGaussian) {
	// Compute the downsampling factor
	M=(int)ROUND(Ts/highTs);

	// Atom profiles for the electron scattering method
	atomProfiles.setup(M,highTs,false);
    }
}

/* Atom description ------------------------------------------------------- */
void Prog_PDBPhantom_Parameters::atomBlobDescription(
    const std::string &_element, double &weight, double &radius) const
{
    int idx = -1;
    weight = radius = 0;
    switch (_element[0])
    {
    case 'H':
        idx = 0;
        break;
    case 'C':
        idx = 1;
        break;
    case 'N':
        idx = 2;
        break;
    case 'O':
        idx = 3;
        break;
    case 'P':
        idx = 4;
        break;
    case 'S':
        idx = 5;
        break;
    case 'F':
        idx = 6;
        break;
    default:
        std::cout << "Unknown :" << _element << std::endl;
        return;
    }
    radius = periodicTable(idx, 0);
    weight = periodicTable(idx, 1);
}

/* Read parameters --------------------------------------------------------- */
void Prog_PDBPhantom_Parameters::read(int argc, char **argv)
{
    fn_pdb = getParameter(argc, argv, "-i");
    fn_out = getParameter(argc, argv, "-o", "");
    if (fn_out == "") fn_out = fn_pdb.without_extension();
    Ts = textToFloat(getParameter(argc, argv, "-sampling_rate", "1"));
    highTs = textToFloat(getParameter(argc, argv, "-high_sampling_rate", "0.08333333"));
    output_dim = textToInteger(getParameter(argc, argv, "-size", "-1"));
    useBlobs = checkParameter(argc, argv, "-blobs");
    usePoorGaussian = checkParameter(argc, argv, "-poor_Gaussian");
}

/* Usage ------------------------------------------------------------------- */
void Prog_PDBPhantom_Parameters::usage()
{
    std::cerr << "convert_pdb2vol\n"
	      << "   -i <pdb file>		       : File to process\n"
	      << "  [-o <fn_root>]		       : Root name for output\n"
	      << "  [-sampling_rate <Ts=1>]	       : Sampling rate (Angstroms/pixel)\n"
	      << "  [-high_sampling_rate <highTs=1/12>]: Sampling rate before downsampling\n"
	      << "  [-size <output_dim>]	       : Final size in pixels (must be a power of 2, if blobs are used)\n"
	      << "  [-blobs]                           : Use blobs instead of scattering factors\n"
              << "  [-poor_Gaussian]                   : Use a simple Gaussian\n"
	      << "\n"
	      << "Example of use: Sample at 1.6A and limit the frequency to 10A\n"
	      << "   xmipp_convert_pdb2vol -i 1o7d.pdb -sampling_rate 1.6\n"
	      << "   xmipp_fourier_filter -i 1o7d.vol -o 1o7d_filtered.vol -low_pass 10 -sampling 1.6 -fourier_mask raised_cosine 0.1\n"
    ;
}

/* Show -------------------------------------------------------------------- */
void Prog_PDBPhantom_Parameters::show()
{
    std::cout << "PDB file:           " << fn_pdb          << std::endl
              << "Sampling rate:      " << Ts              << std::endl
              << "High sampling rate: " << highTs          << std::endl
              << "Size:               " << output_dim      << std::endl
	      << "Use blobs:          " << useBlobs        << std::endl
	      << "Use poor Gaussian:  " << usePoorGaussian << std::endl
    ;
}

/* Compute protein geometry ------------------------------------------------ */
void Prog_PDBPhantom_Parameters::compute_protein_geometry()
{
    Matrix1D<double> limit0(3), limitF(3);
    computePDBgeometry(fn_pdb, centerOfMass, limit0, limitF);
    limit.resize(3);
    XX(limit) = XMIPP_MAX(ABS(XX(limit0)), ABS(XX(limitF)));
    YY(limit) = XMIPP_MAX(ABS(YY(limit0)), ABS(YY(limitF)));
    ZZ(limit) = XMIPP_MAX(ABS(ZZ(limit0)), ABS(ZZ(limitF)));

    // Update output size if necessary
    if (output_dim == -1)
    {
        int max_dim = XMIPP_MAX(CEIL(ZZ(limit) * 2 / Ts) + 5, CEIL(YY(limit) * 2 / Ts) + 5);
        max_dim = XMIPP_MAX(max_dim, CEIL(XX(limit) * 2 / Ts) + 5);
        if (useBlobs)
            output_dim = (int)NEXT_POWER_OF_2(max_dim);
        else 
            output_dim = max_dim+10;
    }
}

/* Create protein at a high sampling rate ---------------------------------- */
void Prog_PDBPhantom_Parameters::create_protein_at_high_sampling_rate()
{
    // Create an empty volume to hold the protein
    int finalDim;
    if (highTs!=Ts)
        finalDim=(int)NEXT_POWER_OF_2(output_dim / (highTs/Ts));
    else
        finalDim=output_dim;
    Vhigh().initZeros(finalDim,finalDim,finalDim);
    Vhigh().setXmippOrigin();
    std::cout << "The highly sampled volume is of size " << XSIZE(Vhigh())
              << std::endl;

    // Fill the volume with the different atoms
    std::ifstream fh_pdb;
    fh_pdb.open(fn_pdb.c_str());
    if (!fh_pdb)
        REPORT_ERROR(1, (std::string)"Prog_PDBPhantom_Parameters::protein_geometry:"
                     "Cannot open " + fn_pdb + " for reading");

    // Process all lines of the file
    while (!fh_pdb.eof())
    {
        // Read an ATOM line
        std::string line;
        getline(fh_pdb, line);
        if (line == "") continue;
        std::string kind = line.substr(0,4);
        if (kind != "ATOM") continue;

        // Extract atom type and position
        // Typical line:
        // ATOM    909  CA  ALA A 161      58.775  31.984 111.803  1.00 34.78
        std::string atom_type = line.substr(13,2);
        double x = textToFloat(line.substr(30,8));
        double y = textToFloat(line.substr(38,8));
        double z = textToFloat(line.substr(46,8));

        // Correct position
        Matrix1D<double> r(3);
        VECTOR_R3(r, x, y, z);
        r -= centerOfMass;
        r /= highTs;

        // Characterize atom
        double weight, radius;
        atomBlobDescription(atom_type, weight, radius);
        blob.radius = radius;
        if (usePoorGaussian) radius=XMIPP_MAX(radius/Ts,4.5);
        double GaussianSigma2=(radius/(3*sqrt(2.0)));
        GaussianSigma2*=GaussianSigma2;
        double GaussianNormalization = 1.0/pow(2*PI*GaussianSigma2,1.5);

        // Find the part of the volume that must be updated
        int k0 = XMIPP_MAX(FLOOR(ZZ(r) - radius), STARTINGZ(Vhigh()));
        int kF = XMIPP_MIN(CEIL(ZZ(r) + radius), FINISHINGZ(Vhigh()));
        int i0 = XMIPP_MAX(FLOOR(YY(r) - radius), STARTINGY(Vhigh()));
        int iF = XMIPP_MIN(CEIL(YY(r) + radius), FINISHINGY(Vhigh()));
        int j0 = XMIPP_MAX(FLOOR(XX(r) - radius), STARTINGX(Vhigh()));
        int jF = XMIPP_MIN(CEIL(XX(r) + radius), FINISHINGX(Vhigh()));

        // Fill the volume with this atom
        for (int k = k0; k <= kF; k++)
            for (int i = i0; i <= iF; i++)
                for (int j = j0; j <= jF; j++)
                {
                    Matrix1D<double> rdiff(3);
                    VECTOR_R3(rdiff, XX(r) - j, YY(r) - i, ZZ(r) - k);
                    rdiff*=highTs;
                    if (useBlobs)
                       Vhigh(k, i, j) += weight * blob_val(rdiff.module(), blob);
                    else if (usePoorGaussian)
                       Vhigh(k, i, j) += weight * 
                          exp(-rdiff.module()*rdiff.module()/(2*GaussianSigma2))*
                          GaussianNormalization;
                }
    }

    // Close file
    fh_pdb.close();
}

/* Create protein at a high sampling rate ---------------------------------- */
void Prog_PDBPhantom_Parameters::create_protein_at_low_sampling_rate()
{
    // Compute the integer downsapling factor
    int M = FLOOR(Ts / highTs);
    double current_Ts = highTs;

    // Use Bsplines pyramid if possible
    int levels = FLOOR(log10((double)M) / log10(2.0) + XMIPP_EQUAL_ACCURACY);
    Vhigh().pyramidReduce(Vlow(), levels);
    current_Ts *= pow(2.0, levels);
    Vhigh.clear();

    // Now scale using Bsplines
    int new_output_dim = CEIL(XSIZE(Vlow()) * current_Ts / Ts);
    Vlow().scaleToSizeBSpline(3, new_output_dim, new_output_dim, new_output_dim,
                                 Vhigh());
    Vlow() = Vhigh();
    Vlow().setXmippOrigin();

    // Return to the desired size
    Vlow().window(FIRST_XMIPP_INDEX(output_dim), FIRST_XMIPP_INDEX(output_dim),
                  FIRST_XMIPP_INDEX(output_dim), LAST_XMIPP_INDEX(output_dim),
                  LAST_XMIPP_INDEX(output_dim), LAST_XMIPP_INDEX(output_dim));
}

/* Blob properties --------------------------------------------------------- */
void Prog_PDBPhantom_Parameters::blob_properties() const
{
    std::ofstream fh_out;
    fh_out.open((fn_out + "_Fourier_profile.txt").c_str());
    if (!fh_out)
        REPORT_ERROR(1, (std::string)"Prog_PDBPhantom_Parameters::blob_properties:"
                     " Cannot open " + fn_out + "_Fourier_profile.txt for output");
    fh_out << "# Freq(1/A) 10*log10(|Blob(f)|^2) Ts=" << highTs << std::endl;
    for (double w = 0; w < 1.0 / (2*highTs); w += 1.0 / (highTs * 500))
    {
        double H = kaiser_Fourier_value(w * highTs, periodicTable(0, 0) / highTs,
                                        blob.alpha, blob.order);
        fh_out << w << " " << 10*log10(H*H) << std::endl;
    }
    fh_out.close();
}

/* Create protein using scattering profiles -------------------------------- */
void Prog_PDBPhantom_Parameters::create_protein_using_scattering_profiles() {
    // Create an empty volume to hold the protein
    Vlow().initZeros(output_dim,output_dim,output_dim);
    Vlow().setXmippOrigin();

    // Fill the volume with the different atoms
    std::ifstream fh_pdb;
    fh_pdb.open(fn_pdb.c_str());
    if (!fh_pdb)
        REPORT_ERROR(1, (std::string)"Prog_PDBPhantom_Parameters::create_protein_using_scattering_profiles:"
                     "Cannot open " + fn_pdb + " for reading");

    // Process all lines of the file
    while (!fh_pdb.eof())
    {
        // Read an ATOM line
        std::string line;
        getline(fh_pdb, line);
        if (line == "") continue;
        std::string kind =line.substr(0,4);
        if (kind != "ATOM") continue;

        // Extract atom type and position
        // Typical line:
        // ATOM    909  CA  ALA A 161      58.775  31.984 111.803  1.00 34.78
        std::string atom_type = line.substr(13,2);
        double x = textToFloat(line.substr(30,8));
        double y = textToFloat(line.substr(38,8));
        double z = textToFloat(line.substr(46,8));

        // Correct position
        Matrix1D<double> r(3);
        VECTOR_R3(r, x, y, z);
        r -= centerOfMass;
        r /= Ts;

        // Characterize atom
	try {
	    double radius=atomProfiles.atomRadius(atom_type[0]);

            // Find the part of the volume that must be updated
            int k0 = XMIPP_MAX(FLOOR(ZZ(r) - radius), STARTINGZ(Vlow()));
            int kF = XMIPP_MIN(CEIL(ZZ(r) + radius), FINISHINGZ(Vlow()));
            int i0 = XMIPP_MAX(FLOOR(YY(r) - radius), STARTINGY(Vlow()));
            int iF = XMIPP_MIN(CEIL(YY(r) + radius), FINISHINGY(Vlow()));
            int j0 = XMIPP_MAX(FLOOR(XX(r) - radius), STARTINGX(Vlow()));
            int jF = XMIPP_MIN(CEIL(XX(r) + radius), FINISHINGX(Vlow()));

            // Fill the volume with this atom
            for (int k = k0; k <= kF; k++)
        	for (int i = i0; i <= iF; i++)
                    for (int j = j0; j <= jF; j++)
                    {
                	Matrix1D<double> rdiff(3);
                	VECTOR_R3(rdiff, XX(r) - j, YY(r) - i, ZZ(r) - k);
			double rdiffModule=rdiff.module();
			if (rdiffModule<radius) 
                	    Vlow(k, i, j) += atomProfiles.volumeAtDistance(
			       atom_type[0],rdiffModule);
                    }
	} catch (Xmipp_error XE) {
	    std::cerr << "Ignoring atom of type *" << atom_type << "*" << std::endl;
	}
    }

    // Close file
    fh_pdb.close();
}

/* Run --------------------------------------------------------------------- */
void Prog_PDBPhantom_Parameters::run()
{
    produceSideInfo();
    compute_protein_geometry();
    if (useBlobs) {
	create_protein_at_high_sampling_rate();
	create_protein_at_low_sampling_rate();
	blob_properties();
    } else if (usePoorGaussian) {
        highTs=Ts;
	create_protein_at_high_sampling_rate();
        Vlow=Vhigh;
        Vhigh.clear();
    } else {
    	create_protein_using_scattering_profiles();
    }
    if (fn_out!="") Vlow.write(fn_out + ".vol");
}
