/***************************************************************************
 *
 * Authors:     Carlos Oscar S. Sorzano (coss@cnb.csic.es)
 *
 * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 * 02111-1307  USA
 *
 *  All comments concerning this program package may be sent to the
 *  e-mail address 'xmipp@cnb.csic.es'
 ***************************************************************************/

#include "phantom_create_micrograph.h"
#include <data/projection.h>

#include <data/args.h>
#include <data/docfile.h>
#include <data/micrograph.h>

/* Empty constructor ------------------------------------------------------- */
Prog_Phantom_Create_Micrograph_Parameters::Prog_Phantom_Create_Micrograph_Parameters()
{
    microscope.command_line=false;
}

/* Read parameters --------------------------------------------------------- */
void Prog_Phantom_Create_Micrograph_Parameters::read(int argc, char **argv)
{
    fn_vol=getParameter(argc,argv,"-vol");
    fn_root=getParameter(argc,argv,"-o","micrograph");
    Xdim=textToInteger(getParameter(argc,argv,"-dim","2048"));
    Nmicrographs=textToInteger(getParameter(argc,argv,"-N","1"));
    density=textToFloat(getParameter(argc,argv,"-density","10"));
    microscope.read(argc,argv);
}

/* Usage ------------------------------------------------------------------- */
void Prog_Phantom_Create_Micrograph_Parameters::usage()
{
    std::cerr << "Usage: xmipp_phantom_create_micrograph\n"
              << "   -vol <Volume>            : Volume for the micrographs\n"
              << "  [-o <rootname=micrograph>]: Rootname for the micrographs\n"
              << "  [-dim <Xdim=2048>]        : Dimension of the micrographs\n"
              << "  [-N <N=1>]                : Number of micrographs\n"
              << "  [-density <d=10>]         : Density of particles (%)\n"
    ;
    microscope.usage();
}

/* Show -------------------------------------------------------------------- */
void Prog_Phantom_Create_Micrograph_Parameters::show()
{
    std::cout
        << "Volume:       " << fn_vol       << std::endl
        << "Rootname:     " << fn_root      << std::endl
        << "Dimension:    " << Xdim         << std::endl
        << "Nmicrographs: " << Nmicrographs << std::endl
        << "Density:      " << density      << std::endl
    ;
    microscope.show();
}

/* Produce side information ------------------------------------------------ */
void Prog_Phantom_Create_Micrograph_Parameters::produce_side_info()
{
    V.read(fn_vol);
    V().setXmippOrigin();
    microscope.Xdim=Xdim;
    microscope.Ydim=Xdim;
    microscope.produce_side_info();
    Nproj=FLOOR(density/100.0*(double)(Xdim*Xdim)/(XSIZE(V())*XSIZE(V())));
}

/* Run -------------------------------------------------------------------.. */
void Prog_Phantom_Create_Micrograph_Parameters::run()
{
    for (int n=0; n<Nmicrographs; n++)
    {
        // Rootname for this micrograph
        FileName dir_micrograph=fn_root+integerToString(n,3);
    
        // Create an empty micrograph ......................................
        system(((std::string)"mkdir "+dir_micrograph).c_str());
        FileName fn_micrograph=dir_micrograph;
        FileName fn_micrograph_clean=fn_micrograph+"_noiseless";
        create_empty_file(dir_micrograph+"/"+fn_micrograph_clean+".raw",
            ((unsigned long long)Xdim)*Xdim);

        std::ofstream fh_inf;
        fh_inf.open((dir_micrograph+"/"+fn_micrograph_clean + ".raw.inf").c_str());
        if (!fh_inf)
            REPORT_ERROR(1, (std::string)"Cannot open " + dir_micrograph+"/"+
                fn_micrograph_clean + ".raw.inf");
        fh_inf << "# Generated by xmipp_phantom_create_micrograph\n";
        fh_inf << "# Image width\n";
        fh_inf << "Xdim= " << Xdim << std::endl;
        fh_inf << "# Image length\n";
        fh_inf << "Ydim= " << Xdim << std::endl;
        fh_inf << "# Pixel depth\n";
        fh_inf << "bitspersample=8\n";
        fh_inf.close();

        Micrograph M;
        M.open_micrograph(dir_micrograph+"/"+fn_micrograph_clean+".raw");
        ImageXmipp Md;
        Md().resize(Xdim,Xdim);
        
        // Create the clean projections ....................................
        DocFile DF_out;
        DF_out.clear();
        DF_out.append_comment("Headerinfo columns: rot (1) , tilt (2), psi (3), Xoff (4), Yoff (5)");
        Matrix1D<double> v(5);
        system(((std::string)"mkdir "+dir_micrograph+"/"+fn_micrograph_clean).c_str());
        FileName fn_proj_clean=fn_micrograph_clean+"/img_noiseless";
        int Xproj=XSIZE(V());
        SelFile SF_out;
        std::cerr << "Creating micrograph " << n << "...\n";
        init_progress_bar(Nproj);
        for (int l=0; l<Nproj; l++)
        {
            // Create projection
            Projection P;
            double rot=rnd_unif(0,360);
            double tilt=rnd_unif(0,180);
            double psi=rnd_unif(0,360);
            project_Volume(V(), P, Xproj, Xproj, rot, tilt, psi);
            FileName fn_image=fn_proj_clean+"_"+integerToString(l,6)+".xmp";
            P.write(dir_micrograph+"/"+fn_image);
            SF_out.insert(fn_image);

            // Store this information in the Docfile
            v(0) = rot;
            v(1) = tilt;
            v(2) = psi;
            v(3) = 0;
            v(4) = 0;
            DF_out.append_comment(fn_image);
            DF_out.append_data_line(v);
            
            // Place this image in the micrograph
            int X=rnd_unif(Xproj,Xdim-Xproj);
            int Y=rnd_unif(Xproj,Xdim-Xproj);
            while (M.search_coord_near(X,Y,Xproj)!=-1)
            {
                X=rnd_unif(Xproj,Xdim-Xproj);
                Y=rnd_unif(Xproj,Xdim-Xproj);
            }
            FOR_ALL_ELEMENTS_IN_MATRIX2D(P())
            {
                int x=X+j;
                int y=Y+i;
                M.set_val(x,y,P(i,j));
                Md(y,x)=P(i,j);
            }
            M.add_coord(X,Y,0,0);
            
            progress_bar(l);
        }
        progress_bar(Nproj);
        DF_out.write(dir_micrograph+"/"+fn_micrograph+".doc");
        M.write_coordinates(0,dir_micrograph+"/"+fn_micrograph+".pos");
        SF_out.write(dir_micrograph+"/"+fn_micrograph_clean+".sel");
        
        // Create the micrograph with noise and CTF ........................
        microscope.apply(Md());
        
        // Write the projections
        system(((std::string)"mkdir "+dir_micrograph+"/"+fn_micrograph).c_str());
        FileName fn_proj=fn_micrograph+"/img";
        DF_out.go_first_data_line();
        SF_out.clear();
        STARTINGY(Md())=STARTINGX(Md())=0;
        for (int l=0; l<Nproj; l++)
        {
            FileName fn_image; fn_image.compose(fn_proj, l, "xmp");
            Particle_coords particle=M.coord(l);
            ImageXmipp I;
            I().resize(Xproj,Xproj);
            I().setXmippOrigin();
            FOR_ALL_ELEMENTS_IN_MATRIX2D(I())
                I(i,j)=Md(particle.Y+i,particle.X+j);
            I.set_rot(DF_out(0));
            I.set_tilt(DF_out(1));
            I.set_psi(DF_out(2));
            I.write(dir_micrograph+"/"+fn_image);
            SF_out.insert(fn_image);
        }
        SF_out.write(dir_micrograph+"/"+fn_micrograph+".sel");
        
        // Write the micrograph
        create_empty_file(dir_micrograph+"/"+fn_micrograph+".raw",
            ((unsigned long long)Xdim)*Xdim*sizeof(float));

        fh_inf.open((dir_micrograph+"/"+fn_micrograph + ".raw.inf").c_str());
        if (!fh_inf)
            REPORT_ERROR(1, (std::string)"Cannot open " + dir_micrograph+"/"+
                fn_micrograph + ".raw.inf");
        fh_inf << "# Generated by xmipp_phantom_create_micrograph\n";
        fh_inf << "# Image width\n";
        fh_inf << "Xdim= " << Xdim << std::endl;
        fh_inf << "# Image length\n";
        fh_inf << "Ydim= " << Xdim << std::endl;
        fh_inf << "# Pixel depth\n";
        fh_inf << "bitspersample=32\n";
        fh_inf.close();

        Micrograph Mctf;
        Mctf.open_micrograph(dir_micrograph+"/"+fn_micrograph+".raw");
        FOR_ALL_ELEMENTS_IN_MATRIX2D(Md())
            Mctf.set_val(j,i,Md(i,j));
        Mctf.close_micrograph();

        // Create the ctfdat
        if (microscope.fn_ctf!="")
        {
            SF_out.go_beginning();
            std::ofstream fh_ctfdat;
            fh_ctfdat.open((dir_micrograph+"/"+fn_micrograph + ".ctfdat").c_str());
            if (!fh_ctfdat)
                REPORT_ERROR(1, (std::string)"Cannot open " + dir_micrograph+"/"+
                    fn_micrograph + ".ctfdat");
            while (!SF_out.eof())
                fh_ctfdat << SF_out.NextImg() << " " << microscope.fn_ctf << std::endl;
            fh_ctfdat.close();
        }

        // Finish with this micrograph
        M.close_micrograph();
    }
}
